
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "euc2unicode.h"

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "game-music-emu-0.5.5/Music_Emu.h"

static u32 TrackNum;

static u8 *pDataBuf;
static u32 DataBufSize;

static Music_Emu *pGME;

// ------------------------------------------------------------------------------------

// ------------------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("Game_Music_Emu 0.5.5\n");
  printf("--------------------\n");
  printf("Author : Shay Green <gblargg@gmail.com>\n");
  printf("Website: http://www.slack.net/~ant/libs/\n");
  printf("Forum  : http://groups.google.com/group/blargg-sound-libs\n");
  printf("License: GNU Lesser General Public License (LGPL)\n");
  printf("\n");
  printf("LibSndGME by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 _TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  {
    FILE *fp=fopen(pFilename,"r");
    if(fp==NULL){
      snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
      return(false);
    }
    
    fseek(fp,0,SEEK_END);
    DataBufSize=ftell(fp);
    fseek(fp,0,SEEK_SET);
    pDataBuf=(u8*)malloc(DataBufSize);
    fread(pDataBuf,DataBufSize,1,fp);
    
    if(fp!=NULL){
      fclose(fp); fp=NULL;
    }
  }
  
  TrackNum=_TrackNum;
  
  u32 SampleRate=44100;
  u32 SamplePerFlame=512;
  
  gme_err_t err=gme_open_data(pDataBuf,DataBufSize,&pGME,SampleRate);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(false);
  }
  
  gme_set_stereo_depth(pGME,1);
  
  err=gme_start_track(pGME,TrackNum);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    gme_delete(pGME);
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(false);
  }
  
  u32 LengthSec=0;
  {
    gme_info_t *pGMEInfo;
    gme_err_t err=gme_track_info(pGME,&pGMEInfo,TrackNum);
    if(err!=NULL){
      snprintf(pState->ErrorMessage,256,"%s",err);
      gme_delete(pGME);
      if(pDataBuf!=NULL){
        free(pDataBuf); pDataBuf=NULL;
      }
      return(false);
    }
    
    LengthSec=pGMEInfo->length;
    if(LengthSec==-1) LengthSec=90;
    
    char msg[1024];
    {
      char *pmsg=msg;
      pmsg[0]=0;
      if(pGMEInfo->game!=NULL) pmsg+=snprintf(pmsg,128,"Game:%s",pGMEInfo->game);
      if(pGMEInfo->song!=NULL) pmsg+=snprintf(pmsg,128,"/ Song:%s",pGMEInfo->song);
      if(pGMEInfo->author!=NULL) pmsg+=snprintf(pmsg,128,"/ Author:%s",pGMEInfo->author);
      if(pGMEInfo->copyright!=NULL) pmsg+=snprintf(pmsg,128,"/ Copyright:%s",pGMEInfo->copyright);
      if(pGMEInfo->comment!=NULL) pmsg+=snprintf(pmsg,128,"/ Comment:%s",pGMEInfo->comment);
      if(pGMEInfo->dumper!=NULL) pmsg+=snprintf(pmsg,128,"/ Dumper:%s",pGMEInfo->dumper);
      if(pGMEInfo->system!=NULL) pmsg+=snprintf(pmsg,128,"/ System:%s",pGMEInfo->system);
    }
    pState->pTitleW=EUC2Unicode_Convert(msg);
    
    gme_free_info(pGMEInfo);
  }
  
  pState->SampleRate=SampleRate;
  pState->SamplesPerFlame=SamplePerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=(float)LengthSec;
  
  LibSnd_DebugOut("GME decoder initialized.\n");
  
  return(true);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  {
    gme_info_t *pGMEInfo;
    gme_err_t err=gme_track_info(pGME,&pGMEInfo,TrackNum);
    if(err!=NULL){
      printf("%s\n",err);
      return;
    }
    
    printf("Track: %d/%d.\n",1+TrackNum,gme_track_count(pGME));
    printf("TotalLength: %d, IntroLength: %d, LoopLength: %d.\n",pGMEInfo->length,pGMEInfo->intro_length,pGMEInfo->loop_length);
    printf("Game: %s.\n",pGMEInfo->game);
    printf("Song: %s.\n",pGMEInfo->song);
    printf("Author: %s.\n",pGMEInfo->author);
    printf("CopyRight: %s.\n",pGMEInfo->copyright);
    printf("Comment: %s.\n",pGMEInfo->comment);
    printf("Dumper: %s.\n",pGMEInfo->dumper);
    printf("System: %s.\n",pGMEInfo->system);
    
    gme_free_info(pGMEInfo);
  }
  
  {
    printf("Tracks info.\n");
    u32 cnt=gme_voice_count(pGME);
    for(u32 idx=0;idx<cnt;idx++){
      const char *pName=gme_voice_name(pGME,idx);
      if(pName!=NULL) printf("Track%d: %s.\n",1+idx,pName);
    }
  }
}

static u32 LibSndInt_Update(u32 *pBufLR);

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(sec<pState->CurrentSec){
    gme_err_t err=gme_start_track(pGME,TrackNum);
    if(err!=NULL){
      printf(pState->ErrorMessage,256,"%s",err);
      return;
    }
    pState->CurrentSec=0;
  }
  
  while(pState->CurrentSec<sec){
    if(LibSndInt_Update(NULL)==0) break;
  }
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(gme_track_ended(pGME)==true) pState->isEnd=true;
  
  if(pState->isEnd==true) return(0);
  
  if(pBufLR==NULL){
    gme_set_stereo_depth(pGME,0);
    pGME->skip(pState->SamplesPerFlame*2);
    gme_set_stereo_depth(pGME,1);
    }else{
    gme_err_t err=gme_play(pGME,pState->SamplesPerFlame*2,(s16*)pBufLR);
    if(err!=NULL){
      snprintf(pState->ErrorMessage,256,"%s",err);
      return(0);
    }
    s16 *pBufLR16=(s16*)pBufLR;
    for(u32 idx=0;idx<pState->SamplesPerFlame;idx++){
      s32 l=pBufLR16[0],r=pBufLR16[1];
      
      // SimpleLPF Normal
      static s32 lastl=0,lastr=0;
      l=(l+lastl)/2; r=(r+lastr)/2;
      lastl=l; lastr=r;
      
      pBufLR16[0]=l; pBufLR16[1]=r;
      pBufLR+=2;
    }
  }
  
  pState->CurrentSec+=(float)pState->SamplesPerFlame/pState->SampleRate;
  
  return(pState->SamplesPerFlame);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  gme_delete(pGME);
}

extern u32 LibSndInt_Helper_GetTracksCount(FILE *fp);
u32 LibSndInt_Helper_GetTracksCount(FILE *fp)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  u8 *pDataBuf;
  u32 DataBufSize;
  
  fseek(fp,0,SEEK_END);
  DataBufSize=ftell(fp);
  fseek(fp,0,SEEK_SET);
  pDataBuf=(u8*)malloc(DataBufSize);
  fread(pDataBuf,DataBufSize,1,fp);
  
  Music_Emu *pGME;
  
  gme_err_t err=gme_open_data(pDataBuf,DataBufSize,&pGME,44100);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(0);
  }
  
  u32 TracksCount=gme_track_count(pGME);
  
  gme_delete(pGME);
  
  if(pDataBuf!=NULL){
    free(pDataBuf); pDataBuf=NULL;
  }
  
  return(TracksCount);
}

TLibSnd_Interface* LibSndGME_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

