
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "liboggTremorRev18100/ivorbiscodec.h"
#include "liboggTremorRev18100/ivorbisfile.h"

//--------------------------------

size_t callbacks_read_func  (void *ptr, size_t size, size_t nmemb, void *datasource)
{
  int len;
  
  len=fread(ptr,size,nmemb,(FILE*)datasource);
  
  return(len);
}

int    callbacks_seek_func  (void *datasource, ogg_int64_t offset, int whence)
{
  u32 pos;
  
  pos=fseek((FILE*)datasource,offset,whence);
  
  return(pos);
}

int    callbacks_close_func (void *datasource)
{
  return(0);
}

long   callbacks_tell_func  (void *datasource)
{
  return(ftell((FILE*)datasource));
}

static const ov_callbacks callbacks = {callbacks_read_func,callbacks_seek_func,callbacks_close_func,callbacks_tell_func};

static const u32 SamplesPerFlame=2048;

typedef struct {
  FILE *FileHandle;
  OggVorbis_File vf;
  u32 SampleRate,ChannelCount;
} TOGGInfo;

static TOGGInfo OGGInfo;

// ------------------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("libogg Tremor Revision 18100 (c)Xiph.org\n");
  printf("LibSndOGG by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  u32 SamplePerFlame=512;
  
  OGGInfo.FileHandle=fopen(pFilename,"r");
  if(OGGInfo.FileHandle==NULL){
    snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
    return(false);
  }
  
  {
    u32 ID;
    fseek(OGGInfo.FileHandle,0,SEEK_SET);
    fread(&ID,1,4,OGGInfo.FileHandle);
    if(ID!=0x004d4b49){
      fseek(OGGInfo.FileHandle,0,SEEK_SET); // normal ogg stream.
      }else{
      fseek(OGGInfo.FileHandle,0x800,SEEK_SET); // skip IKM header.
    }
  }
  
  {
    int ret=ov_open_callbacks((void*)OGGInfo.FileHandle,&OGGInfo.vf,NULL,0,callbacks);
    
    switch(ret){
      case OV_EREAD: LibSnd_DebugOut("ret=OVEREAD A read from media returned an error.\n"); break;
      case OV_ENOTVORBIS: LibSnd_DebugOut("ret=OV_ENOTVORBIS Bitstream is not Vorbis data.\n"); break;
      case OV_EVERSION: LibSnd_DebugOut("ret=OV_EVERSION Vorbis version mismatch.\n"); break;
      case OV_EBADHEADER: LibSnd_DebugOut("ret=OV_EBADHEADER Invalid Vorbis bitstream header.\n"); break;
      case OV_EFAULT: LibSnd_DebugOut("ret=OV_EFAULT Internal logic fault.\n"); break;
    }
    
    if(ret<0){
      snprintf(pState->ErrorMessage,256,"Input does not appear to be an Ogg bitstream.\n");
      return(false);
    }
  }
  
  {
    char **ptr=ov_comment(&OGGInfo.vf,-1)->user_comments;
    
    if(*ptr!=NULL){
      LibSnd_DebugOut("UserComments:\n");
      while(*ptr){
        LibSnd_DebugOut("%s\n",ptr);
        ++ptr;
      }
    }
    
    LibSnd_DebugOut("Encoded by: %s\n",ov_comment(&OGGInfo.vf,-1)->vendor);
    LibSnd_DebugOut("\n");
  }
  
  {
    vorbis_info *vi=ov_info(&OGGInfo.vf,-1);
    
    OGGInfo.SampleRate=vi->rate;
    OGGInfo.ChannelCount=vi->channels;
  }
  
  pState->SampleRate=OGGInfo.SampleRate;
  pState->SamplesPerFlame=SamplesPerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=(float)ov_pcm_total(&OGGInfo.vf,-1)/pState->SampleRate;
  
  {
    vorbis_comment *vc=ov_comment(&OGGInfo.vf,-1);
    if(1<vc->comments){
      if(vc->user_comments[0]!=NULL) pState->pTitleW=Unicode_AllocateCopyFromUTF8(vc->user_comments[0]);
    }
  }
  
  LibSnd_DebugOut("OGG decoder initialized.\n");
  
  return(true);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  vorbis_info *vi=ov_info(&OGGInfo.vf,-1);
  
  printf("SampleRate %dHz\n",OGGInfo.SampleRate);
  printf("ChannelCount %dch\n",OGGInfo.ChannelCount);
  printf("\n");
  if(vi->bitrate_lower!=0) printf("bitrate lower =%dkbps\n",(s32)(vi->bitrate_lower/1000));
  if(vi->bitrate_nominal!=0) printf("bitrate normal=%dkbps\n",(s32)(vi->bitrate_nominal/1000));
  if(vi->bitrate_upper!=0) printf("bitrate upper =%dkbps\n",(s32)(vi->bitrate_upper/1000));
  if(vi->bitrate_window!=0) printf("bitrate window=%dkbps\n",(s32)(vi->bitrate_window/1000));
}

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  ov_pcm_seek(&OGGInfo.vf,(long)(sec*pState->SampleRate));
  pState->CurrentSec=sec;
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pState->isEnd==true) return(0);
  
  u32 unitsize;
  
  if(OGGInfo.ChannelCount==1){
    unitsize=2;
    }else{
    unitsize=4;
  }
  
  const u32 ReadBufSize=SamplesPerFlame;
  u32 ReadBuf[ReadBufSize];
  u32 SamplesCount=0;
  
  {
    int current_section;
    long ret=ov_read(&OGGInfo.vf,(char*)ReadBuf,ReadBufSize*unitsize,&current_section);
    
    if(ret==0){
      LibSnd_DebugOut("Done.\n");
      pState->isEnd=true;
      return(0);
    }
    
    if(ret<0){
      LibSnd_DebugOut("ov_read() decode error:%d. skip.\n",ret);
    }
    
    SamplesCount=ret/unitsize;
  }
  
  if(ReadBufSize<SamplesCount){
    snprintf(pState->ErrorMessage,256,"Fatal error. ReadBufSize=%d, SamplePos=%d.\n",ReadBufSize,SamplesCount);
    return(0);
  }
  
  pState->CurrentSec+=(float)SamplesCount/pState->SampleRate;
  
  if(pBufLR!=NULL){
    if(OGGInfo.ChannelCount==1){
      u16 *pcmbuf=(u16*)ReadBuf;
      for(u32 idx=0;idx<SamplesCount;idx++){
        u16 smp=*pcmbuf++;
        *pBufLR++=smp|(smp<<16);
      }
      }else{
      u32 *pcmbuf=(u32*)ReadBuf;
      for(u32 idx=0;idx<SamplesCount;idx++){
        *pBufLR++=*pcmbuf++;
      }
    }
  }
  
  return(SamplesCount);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  ov_clear(&OGGInfo.vf);
  
  if(OGGInfo.FileHandle!=NULL){
    fclose(OGGInfo.FileHandle); OGGInfo.FileHandle=NULL;
  }
  
  OGGInfo.SampleRate=0;
  OGGInfo.ChannelCount=0;
}

TLibSnd_Interface* LibSndOGG_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

