
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

typedef struct {
  u16 wFormatTag;
  u16 nChannels;
  u32 nSamplesPerSec;
  u32 nAvgBytesPerSec;
  u16 nBlockAlign;
  u16 wBitsPerSample;
  u16 cbSize;
} WAVEFORMATEX;

static FILE *fp;
static WAVEFORMATEX wfex;
static u32 TotalSamplesCount;
static int DataTopOffset;
static int BytePerSample;

static const int SamplesPerFlame=512;
static u32 ReadBuffer[SamplesPerFlame];

// --------------------------------------------------------------------

static bool WaveFile_ReadWaveChunk(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  fseek(fp,0x14,SEEK_SET);
  fread(&wfex,sizeof(wfex),1,fp);
  
  if(wfex.wFormatTag!=0x0001){
    snprintf(pState->ErrorMessage,256,"Illigal CompressFormat Error. wFormatTag=0x%x\n",wfex.wFormatTag);
    return(false);
  }
  
  if((wfex.nChannels!=1)&&(wfex.nChannels!=2)){
    snprintf(pState->ErrorMessage,256,"Channels Error. nChannels=%d\n",wfex.nChannels);
    return(false);
  }
  
  if((wfex.wBitsPerSample!=8)&&(wfex.wBitsPerSample!=16)){
    snprintf(pState->ErrorMessage,256,"Bits/Sample Error. wBitsPerSample=%d\n",wfex.wBitsPerSample);
    return(false);
  }
  
  return(true);
}

static bool WaveFile_SeekDataChunk(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  fseek(fp,0,SEEK_SET);
  
  // find "data"
  {
    char readbuf[256];
    int size=0;
    int ofs=0;
    
    size=fread(readbuf,1,256,fp);
    if(size<4){
      snprintf(pState->ErrorMessage,256,"can not find data chunk.\n");
      return(false);
    }
    
    while(true){
      if(readbuf[ofs]=='d'){
        if((readbuf[ofs+1]=='a')&&(readbuf[ofs+2]=='t')&&(readbuf[ofs+3]=='a')){
          fseek(fp,ofs+4,SEEK_SET);
          break;
        }
      }
      ofs++;
      if(ofs==(size-4)){
        snprintf(pState->ErrorMessage,256,"can not find data chunk.\n");
        return(false);
      }
    }
  }
  
  u32 DataSize;
  fread(&DataSize,4,1,fp);
  
  if(DataSize==0){
    snprintf(pState->ErrorMessage,256,"DataSize is NULL\n");
    return(false);
  }
  
  BytePerSample=0;
  
  if((wfex.nChannels==1)&&(wfex.wBitsPerSample==8)) BytePerSample=1;
  if((wfex.nChannels==2)&&(wfex.wBitsPerSample==8)) BytePerSample=2;
  if((wfex.nChannels==1)&&(wfex.wBitsPerSample==16)) BytePerSample=2;
  if((wfex.nChannels==2)&&(wfex.wBitsPerSample==16)) BytePerSample=4;
  
  if(BytePerSample==0){
    snprintf(pState->ErrorMessage,256,"Illigal Channels or Bits/Sample or no data\n");
    return(false);
  }
  
  TotalSamplesCount=DataSize/BytePerSample;
  
  DataTopOffset=ftell(fp);
  
  return(true);
}

// ------------------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("LibSndWAV by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  fp=fopen(pFilename,"r");
  if(fp==NULL){
    snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
    return(false);
  }
  
  fseek(fp,0,SEEK_SET);
  
  u32 RIFFID;
  fread(&RIFFID,4,1,fp);
  
  if(RIFFID!=0x46464952){ // check "RIFF"
    snprintf(pState->ErrorMessage,256,"no RIFFWAVEFILE error. topdata:0x%04x\n",RIFFID);
    return(false);
  }
  
  if(WaveFile_ReadWaveChunk()==false) return(false);
  if(WaveFile_SeekDataChunk()==false) return(false);
  
  pState->SampleRate=wfex.nSamplesPerSec;
  pState->SamplesPerFlame=SamplesPerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=(float)TotalSamplesCount/pState->SampleRate;
  
  LibSnd_DebugOut("WAV decoder initialized.\n");
  
  return(true);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  LibSnd_DebugOut("fmt:0x%x chns:%d\n",wfex.wFormatTag,wfex.nChannels);
  LibSnd_DebugOut("Smpls/Sec:%d AvgBPS:%d\n",wfex.nSamplesPerSec,wfex.nAvgBytesPerSec);
  LibSnd_DebugOut("BlockAlign:%d Bits/Smpl:%d\n",wfex.nBlockAlign,wfex.wBitsPerSample);
  LibSnd_DebugOut("DataTop:%d\n",DataTopOffset);
}

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  fseek(fp,DataTopOffset+((u32)(sec*pState->SampleRate)*BytePerSample),SEEK_SET);
  pState->CurrentSec=sec;
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pState->isEnd==true) return(0);
  
  int SamplesCount=fread(ReadBuffer,1,SamplesPerFlame*BytePerSample,fp)/BytePerSample;
  if(SamplesCount!=SamplesPerFlame) pState->isEnd=true;
  
  if(wfex.wBitsPerSample==8){
    s8 *readbuf=(s8*)ReadBuffer;
    if(wfex.nChannels==1){ // 8bit mono
      for(u32 idx=0;idx<SamplesCount;idx++){
        s16 s=(((s16)(*readbuf++))-128)<<8;
        *pBufLR++=(s&0xffff)|(s<<16);
      }
      }else{ // 8bit stereo
      for(u32 idx=0;idx<SamplesCount;idx++){
        s16 l=(((s16)(*readbuf++))-128)<<8;
        s16 r=(((s16)(*readbuf++))-128)<<8;
        *pBufLR++=(l&0xffff)|(r<<16);
      }
    }
    }else{
    if(wfex.nChannels==1){ // 16bit mono
      u16 *readbuf=(u16*)ReadBuffer;
      for(u32 idx=0;idx<SamplesCount;idx++){
        u16 s=*readbuf++;
        *pBufLR++=s|(s<<16);
      }
      }else{ // 16bit stereo
      u32 *readbuf=(u32*)ReadBuffer;
      for(u32 idx=0;idx<SamplesCount;idx++){
        *pBufLR++=*readbuf++;
      }
    }
  }
  
  pState->CurrentSec+=(float)SamplesCount/pState->SampleRate;
  
  return(SamplesCount);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(fp!=NULL){
    fclose(fp); fp=NULL;
  }
}

TLibSnd_Interface* LibSndWAV_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

