
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pspuser.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>
#include <psppower.h>
#include <pspctrl.h>

#include "common.h"

#include "GU.h"
#include "VRAMManager.h"
#include "Texture.h"
#include "memtools.h"
#include "unicode.h"
#include "PlayList.h"
#include "LibImg.h"
#include "strtool.h"
#include "CFont.h"
#include "SystemSmallTexFont.h"
#include "sceHelper.h"
#include "euc2unicode.h"

#include "Proc_FileList_BGImg.h"

static void PlayTab_Visible_ON(void);
static void PlayTab_Visible_OFF(void);

#include "Proc_FileList_Folder.h"
#include "Proc_FileList_PlayTab.h"

static char LastBootFolderName[256]="ms0:/MUSIC";

static bool isDirectoryExists(const char *pPath)
{
  if(str_isEmpty(pPath)==true) return(false);
  
  SceUID uid=sceIoDopen(pPath);
  if(uid<0) return(false);
  sceIoDclose(uid);
  return(true);
}

static void Proc_PlayList_CallBack_ChangeState(void)
{
  Folder_RefreshExecutePlayIndex();
  
  if(PlayList_isOpened()==false){
    PlayTab_Visible_OFF();
    }else{
    PlayTab_Visible_ON();
  }
}

static void Proc_Start(EProcState ProcStateLast)
{
  BGImg_Load();
  
  Folder_Init();
  PlayTab_Init();
  
  if(isDirectoryExists(LastBootFolderName)==false) strcpy(LastBootFolderName,BasePath);
  
  Folder_SetPath(LastBootFolderName,NULL);
  
  PlayTab_Visible_ON();
  
  PlayList_CallBack_ChangeState=Proc_PlayList_CallBack_ChangeState;
}

static void Proc_End(void)
{
  strcpy(LastBootFolderName,Folder_GetCurrentPath());
  
  PlayList_CallBack_ChangeState=NULL;
  
  Folder_Free();
  PlayTab_Free();
  
  GuStart();
  BGImg_Draw();
  GuFinish();
  sceGuSync(0,0);
  GuSwapBuffers();
  
  BGImg_Free();
}

static void Proc_KeyDown(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyUp(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyPress(u32 keys,u32 VSyncCount)
{
  if(keys&PSP_CTRL_UP) Folder_MoveCursorUp();
  if(keys&PSP_CTRL_DOWN) Folder_MoveCursorDown();
  
  if(keys&PSP_CTRL_CIRCLE) Folder_ExecuteCurrentFile();
  if(keys&PSP_CTRL_CROSS) PlayList_Stop();
  
  s32 seek=0;
  if(keys&PSP_CTRL_LEFT) seek=-1;
  if(keys&PSP_CTRL_RIGHT) seek=+1;
  if(seek!=0) PlayList_Current_Seek(PlayList_Current_GetCurrentSec()+(seek*5));
  
  s32 vol=0;
  if(keys&PSP_CTRL_TRIANGLE) vol=+1;
  if(keys&PSP_CTRL_SQUARE) vol=-1;
  if(vol!=0){
    s32 lev=PlayList_GetVolume15Max()/32;
    PlayList_SetVolume15(PlayList_GetVolume15()+(vol*lev));
  }
  
  s32 fontsize=0;
  if(keys&PSP_CTRL_LTRIGGER) fontsize=-1;
  if(keys&PSP_CTRL_RTRIGGER) fontsize=+1;
  if(fontsize!=0){
    Folder_AllFreeTexImage();
    Folder_ChangeFontSize(fontsize);
  }
}

static void Proc_KeysUpdate(TKeys *pk,u32 VSyncCount)
{
}

static void Proc_UpdateGU_ins_DrawInfo(void)
{
  static float lastfps=60;
  static float lastclk=0;
  u32 curclk=clock();
  if(lastclk==0){
    lastclk=clock();
    }else{
    u32 repclk=curclk-lastclk;
    lastclk=curclk;
    float sec=((float)repclk)/CLOCKS_PER_SEC;
//    printf("%fmsec.\n",sec*1000);
  }
}

static void Proc_UpdateGU(u32 VSyncCount)
{
  BGImg_VSyncUpdate(VSyncCount);
  BGImg_Draw();
  
  Folder_Update(VSyncCount);
  Folder_Draw();
  
  PlayTab_Update(VSyncCount);
  PlayTab_DrawPanel();
  PlayTab_DrawTitleBar();
  
  Proc_UpdateGU_ins_DrawInfo();
}

static bool Proc_UpdateBlank(void)
{
  return(false);
}

TProc_Interface* Proc_FileList_GetInterface(void)
{
  static TProc_Interface res={
    Proc_Start,
    Proc_End,
    Proc_KeyDown,
    Proc_KeyUp,
    Proc_KeyPress,
    Proc_KeysUpdate,
    Proc_UpdateGU,
    Proc_UpdateBlank,
  };
  return(&res);
}

