
static TTexture FLP_FlameUpTex,FLP_FlameDownTex;
static TTexture FLP_PrgBarNegaTex,FLP_PrgBarPosiTex;
static TTexture FLP_TitleBarNegaTex,FLP_TitleBarPosiTex;
static TTexture FLPI_VolumeTex;
static TTexture FLPI_BatteryChargeTex,FLPI_BatteryLowTex,FLPI_BatteryNormalTex;
static TTexture FLPI_SndPrgTex;

static const s32 DelayMax=20;

#define TitleDelayMaster (60*2)

typedef struct {
  bool Visible;
  s32 UpDelay,DownDelay;
  bool CurrentTop;
  CFont *pCFontTitle;
  u32 TitleBarTop;
  TTexture TitleTex;
  u32 TitleWidth;
  u32 TitleDelay;
  u32 TitleOffset;
} TPlayTab;

static TPlayTab PlayTab;

static void PlayTab_Init(void)
{
  TPlayTab *ppt=&PlayTab;
  
  Texture_CreateFromFile(EVMM_Process,&FLP_FlameUpTex,ETF_RGBA8888,Resources_FileListPath "/FLP_FlameUp.png");
  Texture_CreateFromFile(EVMM_Process,&FLP_FlameDownTex,ETF_RGBA8888,Resources_FileListPath "/FLP_FlameDown.png");
  Texture_CreateFromFile(EVMM_Process,&FLP_PrgBarNegaTex,ETF_RGBA8888,Resources_FileListPath "/FLP_PrgBarNega.png");
  Texture_CreateFromFile(EVMM_Process,&FLP_PrgBarPosiTex,ETF_RGBA8888,Resources_FileListPath "/FLP_PrgBarPosi.png");
  Texture_CreateFromFile(EVMM_Process,&FLP_TitleBarNegaTex,ETF_RGBA8888,Resources_FileListPath "/FLP_TitleBarNega.png");
  Texture_CreateFromFile(EVMM_Process,&FLP_TitleBarPosiTex,ETF_RGBA8888,Resources_FileListPath "/FLP_TitleBarPosi.png");
  
  Texture_CreateFromFile(EVMM_Process,&FLPI_VolumeTex,ETF_RGBA8888,Resources_FileListPath "/FLPI_Volume.png");
  
  Texture_CreateFromFile(EVMM_Process,&FLPI_BatteryChargeTex,ETF_RGBA8888,Resources_FileListPath "/FLPI_BatteryCharge.png");
  Texture_CreateFromFile(EVMM_Process,&FLPI_BatteryLowTex,ETF_RGBA8888,Resources_FileListPath "/FLPI_BatteryLow.png");
  Texture_CreateFromFile(EVMM_Process,&FLPI_BatteryNormalTex,ETF_RGBA8888,Resources_FileListPath "/FLPI_BatteryNormal.png");
  
  Texture_CreateFromFile(EVMM_Process,&FLPI_SndPrgTex,ETF_RGBA8888,Resources_FileListPath "/FLPI_SndPrg.png");
  
  ppt->Visible=false;
  ppt->UpDelay=DelayMax;
  ppt->DownDelay=DelayMax;
  ppt->CurrentTop=true;
  
  ppt->pCFontTitle=pCFont16;
  ppt->TitleBarTop=16;
  Texture_Create(EVMM_Process,&ppt->TitleTex,ETF_RGBA4444,512,ppt->pCFontTitle->GetTextHeight());
  ppt->TitleWidth=0;
  ppt->TitleDelay=0;
  ppt->TitleOffset=0;
  
  Texture_ExecuteSwizzle(&ppt->TitleTex);
}

static void PlayTab_Free(void)
{
  TPlayTab *ppt=&PlayTab;
  
  Texture_Free(EVMM_Process,&FLP_FlameUpTex);
  Texture_Free(EVMM_Process,&FLP_FlameDownTex);
  Texture_Free(EVMM_Process,&FLP_PrgBarNegaTex);
  Texture_Free(EVMM_Process,&FLP_PrgBarPosiTex);
  Texture_Free(EVMM_Process,&FLP_TitleBarNegaTex);
  Texture_Free(EVMM_Process,&FLP_TitleBarPosiTex);
  
  Texture_Free(EVMM_Process,&FLPI_VolumeTex);
  
  Texture_Free(EVMM_Process,&FLPI_BatteryChargeTex);
  Texture_Free(EVMM_Process,&FLPI_BatteryLowTex);
  Texture_Free(EVMM_Process,&FLPI_BatteryNormalTex);
  
  Texture_Free(EVMM_Process,&FLPI_SndPrgTex);
  
  Texture_Free(EVMM_Process,&ppt->TitleTex);
}

static void PlayTab_Visible_OFF(void)
{
  TPlayTab *ppt=&PlayTab;
  
  ppt->Visible=false;
}

static void PlayTab_Visible_ON(void)
{
  TPlayTab *ppt=&PlayTab;
  
  if(PlayList_isOpened()==false){
    PlayTab_Visible_OFF();
    return;
  }
  
  ppt->Visible=true;
  
  TTexture *ptex=&ppt->TitleTex;
  
  Texture_Clear(ptex);
  
  {
    const wchar *pTitleW=PlayList_Current_GetTitleW();
    const u32 pad=16;
    if(Unicode_isEmpty(pTitleW)==true){
      ppt->TitleWidth=pad;
      }else{
      pCFont16->DrawTextW_RGBA4444((u16*)ptex->pImg,ptex->LineSize,0,0,pTitleW);
      ppt->TitleWidth=pCFont16->GetTextWidthW(pTitleW)+pad;
    }
    if(ptex->Width<ppt->TitleWidth) ppt->TitleWidth=ptex->Width;
  }
  
  ppt->TitleDelay=TitleDelayMaster;
  ppt->TitleOffset=0;
}

static void PlayTab_Update(u32 VSyncCount)
{
  TPlayTab *ppt=&PlayTab;
  
  s32 ay=(SelectIndex*TexImageHeight)+(TexImageHeight/2)-(s32)(ShowTop+0.5);
  if(ay<(ScreenHeight/2)){
    ppt->CurrentTop=false;
    }else{
    ppt->CurrentTop=true;
  }
  
  {
    s32 ud=ppt->UpDelay,dd=ppt->DownDelay;
    
    if(ppt->Visible==false){
      ud+=VSyncCount;
      dd+=VSyncCount;
      }else{
      if(ppt->CurrentTop==true){
        ud-=VSyncCount;
        dd+=VSyncCount;
        }else{
        ud+=VSyncCount;
        dd-=VSyncCount;
      }
    }
    
    if(ud<0) ud=0;
    if(DelayMax<ud) ud=DelayMax;
    if(dd<0) dd=0;
    if(DelayMax<dd) dd=DelayMax;
    
    ppt->UpDelay=ud;
    ppt->DownDelay=dd;
  }
  
  if(0<ppt->TitleDelay){
    ppt->TitleDelay--;
    }else{
    ppt->TitleOffset++;
    if(ppt->TitleOffset==(ppt->TitleWidth*2)){
      ppt->TitleDelay=TitleDelayMaster;
      ppt->TitleOffset=0;
    }
  }
  
  if(ppt->Visible==false){
    if(ppt->TitleBarTop<16) ppt->TitleBarTop++;
    }else{
    if(0<ppt->TitleBarTop) ppt->TitleBarTop--;
  }
}

static void PlayTab_Draw_ins_PrgBar(u32 posx,u32 posy,u32 alpha,float per)
{
  TTexture *pntex=&FLP_PrgBarNegaTex,*pptex=&FLP_PrgBarPosiTex;
  
  const u32 w=pntex->Width,h=pntex->Height;
  
  const u32 pbw=w*per;
  
  TRect Rect={-1,-1,-1,-1};
  
  if(0<per){
    Rect.Left=0;
    Rect.Width=pbw;
    Texture_GU_DrawCustom(pptex,posx+Rect.Left,posy,(alpha<<24)|0x00ffffff,Rect);
  }
  
  if(per<100){
    Rect.Left=pbw;
    Rect.Width=w-pbw;
    Texture_GU_DrawCustom(pntex,posx+Rect.Left,posy,(alpha<<24)|0x00ffffff,Rect);
  }
}

static void PlayTab_Draw_ins_TitleBar(u32 posx,u32 posy,u32 alpha,float per)
{
  TTexture *pntex=&FLP_TitleBarNegaTex,*pptex=&FLP_TitleBarPosiTex;
  
  const u32 w=pntex->Width,h=pntex->Height;
  
  const u32 pbw=w*per;
  
  TRect Rect={-1,-1,-1,-1};
  
  if(0<per){
    Rect.Left=0;
    Rect.Width=pbw;
    Texture_GU_DrawCustom(pptex,posx+Rect.Left,posy,(alpha<<24)|0x00ffffff,Rect);
  }
  
  if(per<100){
    Rect.Left=pbw;
    Rect.Width=w-pbw;
    Texture_GU_DrawCustom(pntex,posx+Rect.Left,posy,(alpha<<24)|0x00ffffff,Rect);
  }
}

static float GetPrgPer(void)
{
  float per=0;
  
  float cursec=PlayList_Current_GetCurrentSec(),ttlsec=PlayList_Current_GetTotalSec();
  
  if(ttlsec<cursec) cursec=ttlsec;
  if(ttlsec!=0) per=cursec/ttlsec;
  
  if(per<0) per=0;
  if(1<per) per=1;
  
  return(per);
}

static void PlayTab_DrawPanel(void)
{
  TPlayTab *ppt=&PlayTab;
  
  if((ppt->UpDelay==DelayMax)&&(ppt->DownDelay==DelayMax)) return;
  
  Texture_GU_Start();
  
  if(ppt->UpDelay<DelayMax){
    TTexture *ptex=&FLP_FlameUpTex;
    s32 d=ppt->UpDelay;
    u32 a=0xff-(d*0xff/DelayMax);
    Texture_GU_Draw(ptex,ScreenWidth-ptex->Width+d,0,(a<<24)|0x00ffffff);
  }
  
  if(ppt->DownDelay<DelayMax){
    TTexture *ptex=&FLP_FlameDownTex;
    s32 d=ppt->DownDelay;
    u32 a=0xff-(d*0xff/DelayMax);
    Texture_GU_Draw(ptex,ScreenWidth-ptex->Width+d,ScreenHeight-ptex->Height,(a<<24)|0x00ffffff);
  }
  
  u32 posx,posy,d;
  
  if(ppt->UpDelay<ppt->DownDelay){
    TTexture *ptex=&FLP_FlameUpTex;
    d=ppt->UpDelay;
    posx=ScreenWidth-ptex->Width+d;
    posy=0+8;
    TexFont_DrawText(&SystemSmallTexFont,posx+8,posy,(0xff<<24)|0xffffffff,MSB_VersionFull_UpperCase);
    posy+=4;
    }else{
    TTexture *ptex=&FLP_FlameDownTex;
    d=ppt->DownDelay;
    posx=ScreenWidth-ptex->Width+d;
    posy=ScreenHeight-ptex->Height;
  }
  
  posx+=16; posy+=16;
  
  u32 BarHeight=30;
  
  u32 a=0xff-(d*0xff/DelayMax);
  
  {
    u32 _posx=posx;
    TTexture *picon=&FLPI_VolumeTex;
    Texture_GU_Draw(picon,posx,posy,(a<<24)|0xffffffff);
    _posx+=8+picon->Width;
    float vol=(float)PlayList_GetVolume15()/PlayList_GetVolume15Max();
    PlayTab_Draw_ins_PrgBar(_posx,posy,a,vol);
    _posx+=128;
    {
      char msg[8];
      snprintf(msg,8,"%d%%",(u32)(vol*100));
      _posx-=(SystemSmallTexFont.Width*strlen(msg))+8;
      u32 _posy=posy+((16-SystemSmallTexFont.Height)/2);
      TexFont_DrawText(&SystemSmallTexFont,_posx,_posy,(a<<24)|0xffffffff,msg);
    }
    posy+=BarHeight;
  }
  
  {
    u32 _posx=posx;
    TTexture *picon=NULL;
    if(scePowerIsBatteryCharging()==1){
      picon=&FLPI_BatteryChargeTex;
      }else{
      if(scePowerIsLowBattery()==1){
        picon=&FLPI_BatteryLowTex;
        }else{
        picon=&FLPI_BatteryNormalTex;
      }
    }
    if(picon!=NULL){
      Texture_GU_Draw(picon,_posx,posy,(a<<24)|0xffffffff);
      _posx+=8+picon->Width;
      float per=(float)scePowerGetBatteryLifePercent()/100;
      PlayTab_Draw_ins_PrgBar(_posx,posy,a,per);
      _posx+=128;
      {
        s32 mins=scePowerGetBatteryLifeTime();
        if(mins<0) mins=0;
        char msg[16];
        snprintf(msg,16,"%.2d:%.2d%3d%%",mins/60,mins%60,(u32)(per*100));
        _posx-=(SystemSmallTexFont.Width*strlen(msg))+8;
        u32 _posy=posy+((16-SystemSmallTexFont.Height)/2);
        TexFont_DrawText(&SystemSmallTexFont,_posx,_posy,(a<<24)|0xffffffff,msg);
      }
    }
    posy+=BarHeight;
  }
  
  {
    u32 _posx=posx;
    TTexture *picon=&FLPI_SndPrgTex;
    if(picon!=NULL){
      Texture_GU_Draw(picon,_posx,posy,(a<<24)|0xffffffff);
      _posx+=8+picon->Width;
      
      float per=0;
      float ttlsec=PlayList_Current_GetTotalSec();
      float cursec=PlayList_Current_GetCurrentSec();
      if(ttlsec<cursec) cursec=ttlsec;
      if(ttlsec!=0) per=cursec/ttlsec;
      
      PlayTab_Draw_ins_PrgBar(_posx,posy,a,per);
      _posx+=128;
      {
        char msg[32];
        u32 uttlsec=ttlsec,ucursec=cursec;
        snprintf(msg,32,"%.2d:%.2d / %.2d:%.2d%3d%%",ucursec/60,ucursec%60,uttlsec/60,uttlsec%60,(u32)(per*100));
        _posx-=(SystemSmallTexFont.Width*strlen(msg))+8;
        u32 _posy=posy+((16-SystemSmallTexFont.Height)/2);
        TexFont_DrawText(&SystemSmallTexFont,_posx,_posy,(a<<24)|0xffffffff,msg);
      }
    }
    posy+=BarHeight;
  }
  
  Texture_GU_End();
}

static void PlayTab_DrawTitleBar(void)
{
  TPlayTab *ppt=&PlayTab;
  
  if(ppt->TitleBarTop==16) return;
  
  TTexture *ptex=&FLP_TitleBarPosiTex;
  
  u32 posx=0,posy=ScreenHeight+ppt->TitleBarTop-ptex->Height;
  
  u32 a=0xff*(16-ppt->TitleBarTop)/16;
  
  PlayTab_Draw_ins_TitleBar(posx,posy,a,GetPrgPer());
  
  posx+=12;
  posy+=(ptex->Height-ppt->TitleTex.Height)/2;
  u32 vw=ScreenWidth-posx;
  
  {
    TTexture *ptex=&ppt->TitleTex;
    u32 BaseColor=0x00000000;
    if(ppt->TitleWidth<vw){
      Texture_GU_Draw(ptex,posx,posy,(a<<24)|BaseColor);
      }else{
      u32 ofs=ppt->TitleOffset/2;
      TRect Rect={-1,-1,-1,-1};
      for(u32 idx=1;idx<8;idx++){
        if(idx<=ofs){
          Rect.Left=ofs-idx;
          Rect.Width=1;
          Texture_GU_DrawCustom(ptex,posx-idx,posy,((a*(8-idx)/8)<<24)|BaseColor,Rect);
        }
      }
      Rect.Left=ofs;
      Rect.Width=-1;
      Texture_GU_DrawCustom(ptex,posx,posy,(a<<24)|BaseColor,Rect);
      Texture_GU_Draw(ptex,posx+(ppt->TitleWidth-ofs),posy,(a<<24)|BaseColor);
    }
  }
  
}

