
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>
#include <pspdisplay.h>

#include "common.h"
#include "MemTools.h"

#include "CFont.h"

CFont *pCFont12=NULL,*pCFont14=NULL,*pCFont16=NULL,*pCFont20=NULL,*pCFont24=NULL;

void CFont_Init(ECFontSize CFontSize)
{
  switch(CFontSize){
    case ECFS_12: {
      if(pCFont12!=NULL) return;
      const char *pfn=Resources_FontPath "/Font12.glf";
      printf("Load CFont. [%s]\n",pfn);
      pCFont12=new CFont(pfn);
    } break;
    case ECFS_14: {
      if(pCFont14!=NULL) return;
      const char *pfn=Resources_FontPath "/Font14.glf";
      printf("Load CFont. [%s]\n",pfn);
      pCFont14=new CFont(pfn);
    } break;
    case ECFS_16: {
      if(pCFont16!=NULL) return;
      const char *pfn=Resources_FontPath "/Font16.glf";
      printf("Load CFont. [%s]\n",pfn);
      pCFont16=new CFont(pfn);
    } break;
    case ECFS_20: {
      if(pCFont20!=NULL) return;
      const char *pfn=Resources_FontPath "/Font20.glf";
      printf("Load CFont. [%s]\n",pfn);
      pCFont20=new CFont(pfn);
    } break;
    case ECFS_24: {
      if(pCFont24!=NULL) return;
      const char *pfn=Resources_FontPath "/Font24.glf";
      printf("Load CFont. [%s]\n",pfn);
      pCFont24=new CFont(pfn);
    } break;
  }
}

void CFont_Free(ECFontSize CFontSize)
{
  switch(CFontSize){
    case ECFS_12: {
      if(pCFont12!=NULL){
        delete pCFont12; pCFont12=NULL;
      }
    } break;
    case ECFS_14: {
      if(pCFont14!=NULL){
        delete pCFont14; pCFont14=NULL;
      }
    } break;
    case ECFS_16: {
      if(pCFont16!=NULL){
        delete pCFont16; pCFont16=NULL;
      }
    } break;
    case ECFS_20: {
      if(pCFont20!=NULL){
        delete pCFont20; pCFont20=NULL;
      }
    } break;
    case ECFS_24: {
      if(pCFont24!=NULL){
        delete pCFont24; pCFont24=NULL;
      }
    } break;
  }
}

CFont::CFont(const char *pfn)
{
  fp=fopen(pfn,"r");
  if(fp==NULL){
    printf("CFont: Can not open font file. [%s]\n",pfn);
    SystemHalt();
  }
  
  fread(&Height,4,1,fp);
  
  pWidths=(u8*)safemalloc(0x10000);
  fread(pWidths,0x10000,1,fp);
  pOffsets=(u32*)safemalloc(0x10000*4);
  
  u32 lastofs=4+0x10000;
  for(u32 idx=0;idx<0x10000;idx++){
    u32 w=pWidths[idx];
    if(w==0){
      pOffsets[idx]=0;
      }else{
      u32 size=w*Height;
      pOffsets[idx]=lastofs;
      lastofs+=size;
    }
  }
  
  Caches_LifeCount=1;
  
  for(u32 idx=0;idx<CachesCount;idx++){
    Caches_Unicode[idx]=0;
    Caches_Life[idx]=0;
    Caches_Width[idx]=0;
  }
  
  {
    MaxWidth=0;
    switch(Height){
      case 12: MaxWidth=15; break;
      case 14: MaxWidth=18; break;
      case 16: MaxWidth=20; break;
      case 20: MaxWidth=25; break;
      case 24: MaxWidth=30; break;
      default: {
        printf("CFont: Illigal font height. (%d)\n",Height);
        SystemHalt();
      } break;
    }
    Caches_DataSize=MaxWidth*Height;
    pCaches_Data=(u8*)safemalloc(Caches_DataSize*CachesCount);
  }
}

CFont::~CFont(void)
{
  if(fp!=NULL){
    fclose(fp); fp=NULL;
  }
  
  if(pOffsets!=NULL){
    safefree(pOffsets); pOffsets=NULL;
  }
  
  if(pCaches_Data!=NULL){
    safefree(pCaches_Data); pCaches_Data=NULL;
  }
}

u32 CFont::LoadCache(u16 uidx)
{
  u32 LastAccessCachesIndex=(u32)-1;
  u32 Life=0xffffffff;
  
  for(u32 cidx=0;cidx<CachesCount;cidx++){
    if(Caches_Life[cidx]<Life){
      Life=Caches_Life[cidx];
      LastAccessCachesIndex=cidx;
    }
  }
  
  if(LastAccessCachesIndex==(u32)-1){
    printf("CFont: Can not found last access life cache.\n");
    SystemHalt();
  }
  
  const u32 CachesIndex=LastAccessCachesIndex;
  Caches_Unicode[CachesIndex]=uidx;
  Caches_Life[CachesIndex]=Caches_LifeCount++;
  
  fseek(fp,pOffsets[uidx],SEEK_SET);
  
  u32 Width=pWidths[uidx];
  Caches_Width[CachesIndex]=Width;
  
  if(Width!=0){
    u32 DataSize=MaxWidth*Height;
    u8 *pData=&pCaches_Data[CachesIndex*Caches_DataSize];
    fread(pData,DataSize,1,fp);
    for(u32 idx=0;idx<DataSize;idx++){
      u32 b=(*pData)*0xff/4;
      *pData++=b;
    }
  }
  
  return(CachesIndex);
}

CFont::TCache CFont::GetCache(u16 uidx)
{
  u32 CachesIndex=(u32)-1;
  
  for(u32 cidx=0;cidx<CachesCount;cidx++){
    if(Caches_Unicode[cidx]==uidx){
      CachesIndex=cidx;
      break;
    }
  }
  
  if(CachesIndex==(u32)-1) CachesIndex=LoadCache(uidx);
  
  TCache Cache;
  Cache.pData=&pCaches_Data[CachesIndex*Caches_DataSize];
  Cache.Width=Caches_Width[CachesIndex];
  
  return(Cache);
}

void CFont::DrawChar_White(u32 *pBuf,u32 BufSize,u32 x,u32 y,TCache Cache)
{
  pBuf+=(y*BufSize)+x;
  
  for(u32 y=0;y<Height;y++){
    for(u32 x=0;x<Cache.Width;x++){
      u32 bri=*Cache.pData++;
      if(bri!=0x00){
        u32 col=pBuf[x];
        u32 r=(col>>16)&0xff;
        u32 g=(col>>8)&0xff;
        u32 b=(col>>0)&0xff;
        r=bri+(r*(0x100-bri)/0x100);
        g=bri+(g*(0x100-bri)/0x100);
        b=bri+(b*(0x100-bri)/0x100);
        col=(0xff<<24)|(r<<16)|(g<<8)|(b<<0);
        pBuf[x]=col;
      }
    }
    pBuf+=BufSize;
  }
}

void CFont::DrawChar_RGBA8888(u32 *pBuf,u32 BufSize,u32 x,u32 y,TCache Cache)
{
  pBuf+=(y*BufSize)+x;
  
  for(u32 y=0;y<Height;y++){
    for(u32 x=0;x<Cache.Width;x++){
      u32 bri=(*Cache.pData++)>>0;
      if(bri!=0x00){
        u32 col;
        col=(bri<<24)|(0xff<<16)|(0xff<<8)|(0xff<<0);
        pBuf[x]=col;
      }
    }
    pBuf+=BufSize;
  }
}

void CFont::DrawChar_RGBA4444(u16 *pBuf,u32 BufSize,u32 x,u32 y,TCache Cache)
{
  pBuf+=(y*BufSize)+x;
  
  for(u32 y=0;y<Height;y++){
    for(u32 x=0;x<Cache.Width;x++){
      u32 bri=(*Cache.pData++)>>4;
      if(bri!=0x00){
        u32 col;
        col=(bri<<12)|(0xf<<8)|(0xf<<4)|(0xf<<0);
        pBuf[x]=col;
      }
    }
    pBuf+=BufSize;
  }
}

void CFont::DrawTextA_White(u32 *pBuf,u32 BufSize,u32 x,u32 y,const char *pstr)
{
  while(1){
    char ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_White(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

void CFont::DrawTextW_White(u32 *pBuf,u32 BufSize,u32 x,u32 y,const wchar *pstr)
{
  while(1){
    wchar ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_White(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

void CFont::DrawTextA_RGBA8888(u32 *pBuf,u32 BufSize,u32 x,u32 y,const char *pstr)
{
  while(1){
    char ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_RGBA8888(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

void CFont::DrawTextW_RGBA8888(u32 *pBuf,u32 BufSize,u32 x,u32 y,const wchar *pstr)
{
  while(1){
    wchar ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_RGBA8888(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

void CFont::DrawTextA_RGBA4444(u16 *pBuf,u32 BufSize,u32 x,u32 y,const char *pstr)
{
  while(1){
    char ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_RGBA4444(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

void CFont::DrawTextW_RGBA4444(u16 *pBuf,u32 BufSize,u32 x,u32 y,const wchar *pstr)
{
  while(1){
    wchar ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      if(BufSize<(x+width)) break;
      TCache Cache=GetCache(ch);
      DrawChar_RGBA4444(pBuf,BufSize,x,y,Cache);
      x+=width+1;
    }
  }
}

u32 CFont::GetTextWidthA(const char *pstr) const
{
  u32 w=0;
  
  while(1){
    char ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      w+=width+1;
    }
  }
  
  return(w);
}

u32 CFont::GetTextWidthW(const wchar *pstr) const
{
  u32 w=0;
  
  while(1){
    wchar ch=*pstr++;
    if(ch==0) break;
    u8 width=pWidths[ch];
    if(width!=0){
      w+=width+1;
    }
  }
  
  return(w);
}

u32 CFont::GetTextHeight(void) const
{
  return(Height);
}

const u8* CFont::GetWidthsList(void) const
{
  return(pWidths);
}

