
#include <stdio.h>
#include <stdlib.h>

#include <pspuser.h>

#include "common.h"
#include "memtools.h"
#include "LibSnd.h"
#include "strtool.h"
#include "unicode.h"

#include "PlayList.h"

typedef struct {
  char *pFilename;
} TPlayList_File;

typedef struct {
  u32 FilesCount;
  TPlayList_File *pFiles;
  u32 PlayIndex;
} TPlayList;

static TPlayList PlayList;

void (*PlayList_CallBack_ChangeState)(void);

void PlayList_Init(void)
{
  TPlayList *ppl=&PlayList;
  
  ppl->FilesCount=0;
  ppl->pFiles=NULL;
  ppl->PlayIndex=(u32)-1;
  
  PlayList_CallBack_ChangeState=NULL;
}

void PlayList_Free(void)
{
  PlayList_Stop();
}

bool PlayList_isOpened(void)
{
  TPlayList *ppl=&PlayList;
  
  if(ppl->FilesCount==0) return(false);
  
  return(true);
}

void PlayList_Stop(void)
{
  TPlayList *ppl=&PlayList;
  
  LibSnd_Close();
  
  if(ppl->pFiles!=NULL){
    for(u32 idx=0;idx<ppl->FilesCount;idx++){
      if(ppl->pFiles[idx].pFilename!=NULL){
        safefree(ppl->pFiles[idx].pFilename); ppl->pFiles[idx].pFilename=NULL;
      }
    }
    safefree(ppl->pFiles); ppl->pFiles=NULL;
  }
  
  ppl->FilesCount=0;
  
  ppl->PlayIndex=(u32)-1;
  
  if(PlayList_CallBack_ChangeState!=NULL) PlayList_CallBack_ChangeState();
}

void PlayList_Start(u32 idx)
{
  TPlayList *ppl=&PlayList;
  
  ppl->PlayIndex=idx;
  LibSnd_Start(ppl->pFiles[ppl->PlayIndex].pFilename,0);
  
  if(PlayList_CallBack_ChangeState!=NULL) PlayList_CallBack_ChangeState();
}

bool PlayList_CreateFromPath(const char *pPath,const char *pFilename,u32 TrackNum)
{
  TPlayList *ppl=&PlayList;
  
  PlayList_Stop();
  
  ppl->PlayIndex=(u32)-1;
  
  ppl->FilesCount=0;
  ppl->pFiles=NULL;
  
  {
    SceUID dfd=sceIoDopen(pPath);
    if(dfd<0){
      printf("sceIoDopen error. SEC=0x%08x.\n",dfd);
      SystemHalt();
    }
    
    SceIoDirent dir;
    MemSet8CPU(0,&dir,sizeof(SceIoDirent));
    
    while(1){
      int dreadres=sceIoDread(dfd,&dir);
      if(dreadres==0) break;
      if(dreadres<0){
        printf("sceIoDread error. SEC=0x%08x.\n",dreadres);
        SystemHalt();
      }
      if(FIO_SO_ISDIR(dir.d_stat.st_attr)){
      }
      if(FIO_SO_ISREG(dir.d_stat.st_attr)){
        if(LibSnd_isSupportFile(dir.d_name)==true) ppl->FilesCount++;
      }
    }
    
    sceIoDclose(dfd);
  }
  
  if(ppl->FilesCount==0) return(false);
  
  ppl->pFiles=(TPlayList_File*)safemalloc(ppl->FilesCount*sizeof(TPlayList_File));
  ppl->FilesCount=0;
  
  {
    SceUID dfd=sceIoDopen(pPath);
    if(dfd<0){
      printf("sceIoDopen error. SEC=0x%08x.\n",dfd);
      SystemHalt();
    }
    
    SceIoDirent dir;
    MemSet8CPU(0,&dir,sizeof(SceIoDirent));
    
    while(1){
      int dreadres=sceIoDread(dfd,&dir);
      if(dreadres==0) break;
      if(dreadres<0){
        printf("sceIoDread error. SEC=0x%08x.\n",dreadres);
        SystemHalt();
      }
      if(FIO_SO_ISDIR(dir.d_stat.st_attr)){
      }
      if(FIO_SO_ISREG(dir.d_stat.st_attr)){
        if(LibSnd_isSupportFile(dir.d_name)==true){
          TPlayList_File *pplf=&ppl->pFiles[ppl->FilesCount++];
          pplf->pFilename=MakeFullPath(pPath,dir.d_name);
        }
      }
    }
    
    sceIoDclose(dfd);
  }
  
  if(2<=ppl->FilesCount){
    for(u32 idx1=0;idx1<ppl->FilesCount-1;idx1++){
      for(u32 idx2=idx1+1;idx2<ppl->FilesCount;idx2++){
        const char *pfn1=ppl->pFiles[idx1].pFilename,*pfn2=ppl->pFiles[idx2].pFilename;
        bool sw=false;
        if(isSwapFilename(pfn1,pfn2)==true) sw=true;
        if(sw==true){
          char *tmpf=ppl->pFiles[idx1].pFilename;
          ppl->pFiles[idx1].pFilename=ppl->pFiles[idx2].pFilename;
          ppl->pFiles[idx2].pFilename=tmpf;
        }
      }
    }
  }
  
  {
    u32 playidx=0;
    char *ptagfn=MakeFullPath(pPath,pFilename);
    for(u32 idx=0;idx<ppl->FilesCount;idx++){
      if(isStrEqual(ppl->pFiles[idx].pFilename,ptagfn)==true){
        playidx=idx;
      }
    }
    if(ptagfn!=NULL){
      safefree(ptagfn); ptagfn=NULL;
    }
    PlayList_Start(playidx);
  }
}

void PlayList_Next(void)
{
  TPlayList *ppl=&PlayList;
  
  LibSnd_Close();
  
  u32 playidx=ppl->PlayIndex+1;
  if(playidx==ppl->FilesCount) playidx=0;
  
  PlayList_Start(playidx);
}

void PlayList_Update(void)
{
  if(LibSnd_isOpened()==false) return;
  
  if(LibSnd_GetisEnd()==true){
    PlayList_Next();
  }
}

bool PlayList_isSupportFile(const char *pFilename)
{
  return(LibSnd_isSupportFile(pFilename));
}

u32 PlayList_GetVolume15Max(void)
{
  return(LibSnd_GetVolume15Max());
}

u32 PlayList_GetVolume15(void)
{
  return(LibSnd_GetVolume15());
}

void PlayList_SetVolume15(s32 Volume15)
{
  LibSnd_SetVolume15(Volume15);
}

void PlayList_Current_Seek(float sec)
{
  if(LibSnd_isOpened()==false) return;
  
  LibSnd_Seek(sec);
}

float PlayList_Current_GetTotalSec(void)
{
  if(LibSnd_isOpened()==false) return(0);
  
  const TLibSndConst_State *pState=LibSnd_GetState();
  return(pState->TotalSec);
}

float PlayList_Current_GetCurrentSec(void)
{
  if(LibSnd_isOpened()==false) return(0);
  
  const TLibSndConst_State *pState=LibSnd_GetState();
  return(pState->CurrentSec);
}

const wchar* PlayList_Current_GetTitleW(void)
{
  if(LibSnd_isOpened()==false) return(NULL);
  
  const TLibSndConst_State *pState=LibSnd_GetState();
  return(pState->pTitleW);
}

const char* PlayList_Current_GetFilename(void)
{
  if(LibSnd_isOpened()==false) return(NULL);
  
  TPlayList *ppl=&PlayList;
  
  return(ppl->pFiles[ppl->PlayIndex].pFilename);
}

